const coral = @import("coral");

const gfx = @import("../gfx.zig");

const std = @import("std");

pub fn load_bmp(arena: *std.heap.ArenaAllocator, path: []const u8) gfx.Assets.Format.Error!gfx.Desc {
	const header = try coral.files.bundle.read_little(path, 0, extern struct {
		type: [2]u8 align (1),
		file_size: u32 align (1),
		reserved: [2]u16 align (1),
		image_offset: u32 align (1),
		header_size: u32 align (1),
		pixel_width: i32 align (1),
		pixel_height: i32 align (1),
		color_planes: u16 align (1),
		bits_per_pixel: u16 align (1),
		compression_method: u32 align (1),
		image_size: u32 align(1),
		pixels_per_meter_x: i32 align (1),
		pixels_per_meter_y: i32 align (1),
		palette_colors_used: u32 align (1),
		important_colors_used: u32 align (1),
	}) orelse {
		return error.Unsupported;
	};

	if (!std.mem.eql(u8, &header.type, "BM")) {
		return error.Unsupported;
	}

	const pixel_width = std.math.cast(u16, header.pixel_width) orelse {
		return error.Unsupported;
	};

	const pixels = try arena.allocator().alloc(coral.io.Byte, header.image_size);
	const bytes_per_pixel = header.bits_per_pixel / coral.io.bits_per_byte;
	const alignment = 4;
	const byte_stride = pixel_width * bytes_per_pixel;
	const padded_byte_stride = alignment * @divTrunc((byte_stride + alignment - 1), alignment);
	const byte_padding = coral.scalars.sub(padded_byte_stride, byte_stride) orelse 0;
	var buffer_offset: usize = 0;
	var file_offset = @as(usize, header.image_offset);

	while (buffer_offset < pixels.len) {
		const line = pixels[buffer_offset .. buffer_offset + byte_stride];

		if (try coral.files.bundle.read_bytes(path, file_offset, line) != byte_stride) {
			return error.Unsupported;
		}

		file_offset = line.len + byte_padding;
		buffer_offset += padded_byte_stride;
	}

	return .{
		.texture = .{
			.format = switch (header.bits_per_pixel) {
				24 => .bgr888,
				32 => .bgra8888,
				else => return error.Unsupported,
			},

			.width = pixel_width,
			.data = pixels,
			.access = .static,
		}
	};
}
